/*
  GNUJSP - a free JSP1.0 implementation

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package org.gjt.jsp.jsdk20;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.HttpURLConnection;
import javax.servlet.GenericServlet;
import javax.servlet.RequestDispatcher;
import javax.servlet.Servlet;
import javax.servlet.ServletContext;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.ServletException;
import javax.servlet.SingleThreadModel;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.jsp.JspWriter;
import org.gjt.jsp.JspConfig;
import org.gjt.jsp.JspMsg;
import org.gjt.jsp.JspServlet;

/**
 * Provides an implementation of RequestDispatcher as mandated by
 * Servlet API 2.1.  This is used by ServletContextWrapper in attempt
 * to provide support for jsp:include and jsp:forward when the underlying
 * servlet engine only supports the 2.0 API.
 *
 * @author Wes Biggs <wes@gjt.org>, Carsten Heyl <alph@gjt.org>
 */
public class RequestDispatcherImpl implements RequestDispatcher, JspMsg {
    private ServletContext sc;
    private String path;
    private String query;
    private JspServlet gnujsp = null;
    private JspWriter jspWriter = null;
    private boolean pathDebug;
    private boolean debug;
    private boolean useHTTPFallback;

    /**
     * helper for string printing
     */
    private static String qq(String s) {
	if(s == null) {
	    return "null";
	} else {
	    return "'"+s+"'";
	}
    }

    /** Invoked by ServletContextWrapper.getResourceDispatcher(path). */
    RequestDispatcherImpl(ServletContext sc,
			  String path,
			  JspWriter jspWriter,
			  JspServlet gnujsp) {
	this.sc = sc;
	this.path = path;
	this.gnujsp = gnujsp;
	this.jspWriter = jspWriter;

	pathDebug = gnujsp.getPathDebug();
	debug = gnujsp.getDebug();
	useHTTPFallback = gnujsp.getUseHTTPFallback();

	// Check for name/value parameters
	int p = path.indexOf('?');
	if (p != -1) {
	    query = path.substring(p + 1);
	    this.path = path.substring(0, p);
	}
    }

    public void forward(ServletRequest srequest, ServletResponse sresponse) 
	throws ServletException, IOException
    {
	String s = null;
	    
	HttpServletRequest request = (HttpServletRequest) srequest;
	HttpServletRequestWrapper wrequest = new HttpServletRequestWrapper(request, query);

	HttpServletResponse response = (HttpServletResponse) sresponse;

	// We set parameters similiar to include

	wrequest.setAttribute("javax.servlet.forward.request", 
			      path);

	/* servletpath component FIXME should be for wrequest */ 
	s = request.getServletPath();
	if(s != null) {
	    wrequest.setAttribute("javax.servlet.forward.servlet", s);
	}
	/* the pathinfo component FIXME should be for wrequest */
	wrequest.setAttribute("javax.servlet.forward.path",
			      path);
	/* the querystring */
	if (query != null) {
	    wrequest.setAttribute("javax.servlet.forward.query", query);
	}

	String jspURI = wrequest.getRequestURI();

	// May use that if ServletPath is fixed in Wrapper
	//	String jspURI = requestToJspURI(wrequest);

	if(jspURI == null) {
	    throw new FileNotFoundException
		(JspConfig.getLocalizedMsg
		 (ERR_sp10_2_13_5_could_not_find_forward_file)
		 + ": " + path);
	}
	try {
	    gnujsp.xservice(wrequest, response, path);
	} catch(FileNotFoundException e) {
	    // FIXME: another way without exception to check if jsp 
	    // file exists?

	    // the file was not found as a jsp page, 
	    // so we try it as a servlet

	    if(!doServletInclude(jspURI,
				 request,
				 response)) {
		// the file was not found as a jsp page or servlet,
		// This is not up to spec, but works for some cases
		if (query != null) path += "?" + query;
		((HttpServletResponse) response).sendRedirect(path);
	    }
	} catch(Exception e) {
	    // IllegalAccess, ClassNotFound, Instantiation
	    // FIXME: reason? how to handle?
	    if(debug) {
		gnujsp.log("doForward: exception", e);
	    }
	    throw new ServletException(e.toString());
	}
    }

    public void include(ServletRequest srequest, ServletResponse sresponse)
    	throws ServletException, IOException
    {
	String s = null;
	if(path == null) 
	    throw new IOException
		(JspConfig.getLocalizedMsg(ERR_gnujsp_internal_error)
		 + ": path==null");
	    
	// B.9.2 Additional attributes for include
	// Should they be set on request? on wrapped request?
	// We can't remove them after include, so I guess
	// it's a wrapped request. (alph)

	HttpServletRequest request = (HttpServletRequest) srequest;
	HttpServletRequestWrapper wrequest = new HttpServletRequestWrapper(request, query);

	HttpServletResponse response = (HttpServletResponse) sresponse;

	wrequest.setAttribute("javax.servlet.include.request", 
			      path);

	/* servletpath component FIXME should be for wrapped request */ 
	s = request.getServletPath();
	if(s != null) {
	    wrequest.setAttribute("javax.servlet.include.servlet", s);
	}
	/* the pathinfo component */
	wrequest.setAttribute("javax.servlet.include.path",
			      path);
	/* the querystring */
	if (query != null) {
	    wrequest.setAttribute("javax.servlet.include.query", query);
	}

	String jspURI = wrequest.getRequestURI();

	if(pathDebug) {
	    gnujsp.log("include.pathdebug.uri1="+qq(jspURI));
	}

	// May use that if ServletPath is fixed in Wrapper
	//	String jspURI = requestToJspURI(wrequest);

	if(jspURI == null) {
	    throw new FileNotFoundException
		(JspConfig.getLocalizedMsg
		 (ERR_sp10_2_13_5_could_not_find_include_file)
		 + ": " + path);
	}
	try {
	    JspWriter w = new JspWriterImpl(jspWriter, 
					    8192 /* FIXME */, 
					    false);
	    HttpServletResponseWrapper wres = 
		new HttpServletResponseWrapper(response);
	    PrintWriter pw = new PrintWriter(w); 
	    wres.setWriter(pw);
	    gnujsp.xservice(wrequest, wres, jspURI);
	    pw.flush();
	} catch(FileNotFoundException e) {
	    // FIXME: another way without exception to check if jsp 
	    // file exists?

	    // the file was not found as a jsp page, 
	    // so we try it as a servlet

	    if(!doServletInclude(jspURI,
				 wrequest,
				 response)) {
		// the file was not found as a jsp page or servlet,
		// so we try to get it via HTTP protocol
		if(useHTTPFallback) {
		    doHTTPInclude(jspWriter, wrequest);
		} else {
		    throw e;
		}
	    }
	} catch(Exception e) {
	    // IllegalAccess, ClassNotFound, Instantiation
	    // FIXME: reason? Which to pass?
	    if(debug) {
		gnujsp.log("doInclude: exception", e);
	    }
	    throw new ServletException(e.toString());
	}
    }

    private boolean doServletInclude(String jspURI,
				  HttpServletRequest request,
				  HttpServletResponse res) 
	throws IOException, ServletException
    {
	String servletName = jspURI.substring(jspURI.lastIndexOf("/")+1);
	Servlet servlet = sc.getServlet(servletName);
	if(servlet != null) {
	    // we need a response wrapper to that
	    // the servlet may use res.getServletOutputStream()
	    // without getting java.lang.IllegalStateException
	    // because getWriter() was already called.
	    // I didn't used it in doInclude() because it
	    // is not needed and may break things (not tested) (alph)
	    HttpServletResponse response = 
		new HttpServletResponseWrapper(res);
	    
	    
	    // FIXME: calling init must be done by servlet engine!
	    // FIXME: we don't know if it is already called!
	    // FIXME: document that! (alph)
	    // servlet.init(info);
	    
	    // This is from ApacheJSSI:
	    
	    // there is no defined way to get a STM-servlet to execute
	    // it. The behaviour is undefined
	    if (servlet instanceof SingleThreadModel &&
		servlet instanceof GenericServlet) {
		((GenericServlet) servlet).log 
		    (JspConfig.getLocalizedMsg
		     (ERR_gnujsp_unsafe_singletread_execute));
	    }

	    // execute servlet
	    servlet.service((ServletRequest) request, 
			    (ServletResponse) response);
	}

	return (servlet != null);
    }
	    
    private void doHTTPInclude(JspWriter out, 
			       HttpServletRequest request) 
	throws IOException
    {
	HttpURLConnection conn = null;
	BufferedReader r = null;

	int resCode = 0;
	try {
	    char buf[] = new char[8192];
	    URL url = new URL(makeRequestURL(request));
		
	    conn = (HttpURLConnection) url.openConnection();
	    r = new BufferedReader(new InputStreamReader(conn.getInputStream()));

	    int i = 0;
	    while((i = r.read(buf, 0, buf.length)) > -1) {
		if(i > 0) {
		    out.write(buf, 0, i);
		}
	    }
	    resCode = conn.getResponseCode();
	} catch(MalformedURLException ee) {
	    throw new FileNotFoundException(path);
	} finally {
	    if(r != null) {
		try { r.close(); } catch(Exception ignore) {}
	    }
	    if(conn != null) {
		try { conn.disconnect(); } catch(Exception ignore) {}
	    }
	}
	if(resCode == HttpServletResponse.SC_NOT_FOUND ) {
	    throw new FileNotFoundException(resCode + ": "+ path);
	}
    }

    /**
     *
     * reconstruct the url this page was called with
     *
     */ 
    private String makeRequestURL(HttpServletRequest req) 
	throws IOException
    {
	int port = req.getServerPort();
	String strPort = "";
	if(port != 80)
		strPort = ":" + port;

	String path = req.getPathInfo();
	if(path == null || path.equals(""))
	    path = req.getServletPath();

        if(path == null)	
	    throw new IOException(JspConfig.getLocalizedMsg
				  (ERR_gnujsp_could_not_reconstruct_url));
          
	return req.getScheme()+"://"+req.getServerName()+strPort+path;        
    }
}
